/*
    SPDX-FileCopyrightText: 2020 David Edmundson <davidedmundson@kde.org>
    SPDX-FileCopyrightText: 2021 David Redondo <kde@david-redondo.de>

    SPDX-License-Identifier: LGPL-2.1-only OR LGPL-3.0-only OR LicenseRef-KDE-Accepted-LGPL
*/

#include "abstract_drop_handler.h"

namespace KWin
{
AbstractDropHandler::AbstractDropHandler(QObject *parent)
    : QObject(parent)
{
}

}

#include "moc_abstract_drop_handler.cpp"
