/*
    SPDX-FileCopyrightText: 2020 Vlad Zahorodnii <vlad.zahorodnii@kde.org>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "utils/vsyncmonitor.h"

namespace KWin
{

VsyncMonitor::VsyncMonitor() = default;

} // namespace KWin

#include "moc_vsyncmonitor.cpp"
