/*
    SPDX-FileCopyrightText: 2023 Vlad Zahorodnii <vlad.zahorodnii@kde.org>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "core/graphicsbufferallocator.h"

namespace KWin
{

GraphicsBufferAllocator::GraphicsBufferAllocator()
{
}

GraphicsBufferAllocator::~GraphicsBufferAllocator()
{
}

} // namespace KWin
